import functools
import logging
import sqlite3

from ._abstract import Artifact

logger_log = logging.getLogger("xleapp.logfile")


def core_artifact(cls: Artifact):
    """Decorator to mark a core artifact

    Args:
        cls: Artifact class object

    Raises:
        AttributeError

    """

    @functools.wraps(cls)
    def core_wrapper(cls):
        if issubclass(cls, Artifact):
            cls.core = True
            cls.selected = True
            cls.report = False
            return cls
        else:
            raise AttributeError(
                f"Class object {str(cls)} is not an Artifact! "
                f'Error setting property "core_artifact" on class!',
            )

    return core_wrapper(cls)


def long_running_process(cls: Artifact):
    """Decorator to mark an artifact as long running

    Args:
        cls: Artifact class object

    Raises:
        AttributeError

    Returns:
        Artifact: Artifact is marked as long running process
    """

    @functools.wraps(cls)
    def lrp_wrapper(cls):
        if issubclass(cls, Artifact):
            cls.long_running_process = True
            return cls
        else:
            raise AttributeError(
                f"Class object {str(cls)} is not an Artifact! "
                f'Error setting property "long_running_process" on class!',
            )

    return lrp_wrapper(cls)


class Search:
    def __init__(
        self,
        *args,
        file_names_only: bool = False,
        return_on_first_hit: bool = True,
    ):
        self.return_on_first_hit = return_on_first_hit
        self.file_names_only = file_names_only
        self.search = [*args]

    def __call__(self, func):
        def search_wrapper(cls) -> bool:
            try:
                with cls.context(
                    regex=self.search,
                    file_names_only=self.file_names_only,
                    return_on_first_hit=self.return_on_first_hit,
                ) as artifact:
                    func(artifact)
                    cls.processed = True
            except sqlite3.OperationalError as ex:
                logger_log.error(f"-> Error {ex}")
            return cls.processed

        functools.update_wrapper(search_wrapper, func)
        return search_wrapper



# -*- coding: utf-8 -*-
"""
xLEAPP 
"""

from ._authors import __authors__, __contributors__
from ._version import __project__, __version__
from .artifacts import Artifact, Search, core_artifact, long_running_process
from .helpers.db import open_sqlite_db_readonly
from .helpers.decorators import timed
from .report import WebIcon
from .templating import ArtifactHtmlReport, Template

VERSION = f"{__project__} v{__version__}"



import logging
import shutil
import typing as t
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from textwrap import TextWrapper

import prettytable

import xleapp.helpers.utils as utils

from ._abstract import Artifact
from ._decorators import Search, core_artifact, long_running_process

if t.TYPE_CHECKING:
    from xleapp.app import XLEAPP

logger_log = logging.getLogger("xleapp.logfile")


def crunch_artifacts(app: "XLEAPP") -> None:
    for artifact in app.artifacts:
        if not artifact.value.selected:
            continue
        # Now ready to run
        # Special processing for iTunesBackup Info.plist as it is a
        # separate entity, not part of the Manifest.db. Seeker won't find it
        if artifact.cls_name == "ITunesBackupInfo":
            info_plist_path = Path(app.input_path) / "Info.plist"
            if info_plist_path.exists():
                app.artifacts.process_artifact(artifact)
            else:
                logger_log.info("Info.plist not found for iTunes Backup!")
            # noqa GuiWindow.SetProgressBar(categories_searched * ratio)
        else:
            app.artifacts.process_artifact(artifact)


def generate_artifact_path_list(artifacts) -> None:
    """Generates path file for usage with Autopsy

    Args:
        artifacts(list): List of artifacts to get regex from.
    """
    logger_log.info("Artifact path list generation started.")

    with open("path_list.txt", "w") as paths:
        regex_list = []
        for artifact in artifacts:
            if isinstance(artifact.search_dirs, tuple):
                [regex_list.append(item) for item in artifact.search_dirs]
            else:
                regex_list.append(artifact.search_dirs)
        # Create a single list removing duplications
        ordered_regex_list = "\n".join(set(regex_list))
        logger_log.info(ordered_regex_list)
        paths.write(ordered_regex_list)

        logger_log.info("Artifact path list generation completed")


def generate_artifact_table(artifacts) -> None:
    """Generates artifact list table.

    Args:
        artifacts(list): List of artifacts to get regex from.
    """
    headers = ["Short Name", "Full Name", "Search Regex"]
    wrapper = TextWrapper(expand_tabs=False, replace_whitespace=False, width=60)
    output_table = prettytable.PrettyTable(headers, align="l")
    output_table.hrules = prettytable.ALL
    output_file = Path("artifact_table.txt")

    logger_log.info("Artifact table generation started.")

    with open(output_file, "w") as paths:
        for key, value in artifacts:
            short_name = key
            full_name = value.cls.name
            search_regex = value.cls.search_dirs
            if isinstance(search_regex, tuple):
                search_regex = "\n".join(search_regex)
            output_table.add_row([short_name, full_name, wrapper.fill(search_regex)])
        paths.write(output_table.get_string(title="Artifact List", sortby="Short Name"))
    logger_log.info(f"Table saved to: {output_file}")
    logger_log.info("Artifact table generation completed")


def copyfile(
    report_folder: Path, name: str, input_file: t.Union[PathLike, str], output_file: str
) -> Path:
    """Exports file to report folder

    File will be located under report_folder\\export\\artifact_class

    Args:
        names(str): name of the artifact class
        input_file (str): input file name/path
        output_file (str): output file name

    Returns:
        output_file (Path): Path object of the file save location and name.
    """
    report_folder = report_folder
    artifact_folder = name
    output_folder = Path(report_folder / "export" / artifact_folder)
    output_folder.mkdir(parents=True, exist_ok=True)

    if utils.is_platform_windows():
        input_file = Path(f"\\\\?\\{input_file.resolve()}")

    output_file = Path(output_folder / output_file)
    shutil.copy2(input_file, output_file)
    logger_log.debug(f"File {input_file.name} copied to " f"{output_file}")
    return output_file


def filter_artifacts(l: t.Iterator[Artifact]):
    for artifact in l:
        if getattr(artifact, "selected", False) == True:
            yield artifact



# -*- coding: utf-8 -*-
"""[summary]
"""

import codecs
import csv
import sqlite3
from abc import abstractmethod
from pathlib import Path
import typing as t

import simplekml
from xleapp import report


class Options:
    def __set_name__(self, owner, name) -> None:
        self.name = str(name)

    def __get__(self, obj, type=None) -> dict:
        return obj.__dict__.get(self.name) or dict()

    def __set__(self, obj, options) -> None:
        if not isinstance(options, dict):
            raise TypeError(f"{self.name!r} is {type(options)!r} instead of {dict!r}!")
        obj.__dict__[self.name] = options["options"]
        for option, val in options["options"].items():
            if option in ["file", "data_list", "data_headers"]:
                obj.__dict__[option] = val


class DBManager:
    connection = None
    report_folder: Path = None
    file: Path = None
    options: dict = Options()

    def __init__(self, report_folder: Path, db_folder: Path, options: dict) -> None:
        self.report_folder = report_folder / db_folder
        self.report_folder.mkdir(parents=True, exist_ok=True)
        self.options = options

    def __enter__(self) -> "DBManager":
        self.connection = sqlite3.connect(self.file, isolation_level="exclusive")
        self.connection.row_factory = sqlite3.Row
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self.connection.commit()

    @abstractmethod
    def save(self, *args) -> None:
        """Saves files to database

        Returns:
            None
        """
        NotImplementedError(f"{self.__name__} requires a `save()` function!")

    @abstractmethod
    def _create(self) -> None:
        """Creates database file

        Returns:
            None
        """
        NotImplementedError(f"{self.__name__} requires a `_create()` function!")


class KmlDBManager(DBManager):
    def __init__(self, report_folder: Path, **options) -> None:
        super().__init__(
            report_folder=report_folder, db_folder="_KML_Exports", options=options
        )
        self.file = self.report_folder / "_latlong.db"

        if not self.file.exists():
            db = sqlite3.connect(self.file, isolation_level="exclusive")
            cursor = db.cursor()
            cursor.execute(
                """
                CREATE TABLE data(key TEXT, latitude TEXT, longitude TEXT, activity TEXT)
                """,
            )
            db.commit()

    def save(self) -> None:
        kml = simplekml.Kml(open=1)

        with KmlDBManager(self.report_folder) as db:
            db.connection.execute("""PRAGMA synchronous = EXTRA""")
            db.connection.execute("""PRAGMA journal_mode = WAL""")
            db.connection.commit()

            for row in self.data_list:
                modifiedDict = dict(zip(self.data_headers, row))

                times = modifiedDict["Timestamp"]
                lon = modifiedDict["Longitude"]
                lat = modifiedDict["Latitude"]

                if lat:
                    pnt = kml.newpoint()
                    pnt.name = times
                    pnt.description = f"Timestamp: {times} - {self.name}"
                    pnt.coords = [(lon, lat)]

                    db.connection.execute(
                        "INSERT INTO data VALUES(?,?,?,?)",
                        (times, lat, lon, self.name),
                    )

            kml.save(db.report_folder / f"{self.name}.kml")


class TimelineDBManager(DBManager):
    def __init__(self, report_folder: Path, **options):
        super().__init__(
            report_folder=report_folder, db_folder="_Timeline", options=options
        )
        self.file = self.report_folder / "t1.db"

        if not self.file.exists():
            db = sqlite3.connect(self.file, isolation_level="exclusive")
            cursor = db.cursor()
            cursor.execute(
                """
                CREATE TABLE data(key TEXT, activity TEXT, datalist TEXT)
                """,
            )
            db.commit()

    def save(self) -> None:
        with TimelineDBManager(self.report_folder) as db:
            db.connection.execute("""PRAGMA synchronous = EXTRA""")
            db.connection.execute("""PRAGMA journal_mode = WAL""")

            for row in self.data_list:
                modifiedlist = list(
                    map(lambda x, y: x.upper() + ": " + str(y), self.data_headers, row),
                )
                db.connection.executemany(
                    "INSERT INTO data VALUES(?,?,?)",
                    [(str(row[0]), self.name.upper(), str(modifiedlist))],
                )


class TsvManager(DBManager):
    def __init__(self, report_folder: Path, **options):
        super().__init__(
            report_folder=report_folder, db_folder="_TSV Exports", options=options
        )
        self.file = f"{self.file}.tsv"

    def __enter__(self):
        self.connection = codecs.open(self.file, "a", "utf-8-sig")
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self.connection.close()

    def save(self):

        with self as file:
            tsv_writer = csv.writer(file, delimiter="\t")
            tsv_writer.writerow(self.data_headers)
            for i in self.data_list:
                tsv_writer.writerow(i)



import datetime
import logging
import os
import typing as t
from collections import UserDict
from enum import Enum
from functools import cached_property
from pathlib import Path

import jinja2
from jinja2 import Environment

import xleapp.artifacts as artifacts
import xleapp.templating as templating
from xleapp.helpers.descriptors import Validator

from ._version import __project__, __version__
from .artifacts.services import Artifacts
from .helpers.search import FileSeekerBase
from .templating._ext import IncludeLogFileExtension

logger_log = logging.getLogger("xleapp.logfile")


class Device(UserDict):
    def table(self):
        return [[key, value] for key, value in self.data.items()]


class OutputFolder(Validator):
    def validator(self, value):
        if not isinstance(value, (str, Path, os.PathLike)):
            raise TypeError(f"Expected {value!r} to be one of: str, Path, or Pathlike!")
        if not Path(value).exists():
            raise FileNotFoundError(f"{value!r} must already exists!")


class XLEAPP:

    debug: bool = False

    project: str

    version: str

    device: Device = Device()

    default_configs: dict[str, t.Any] = dict()

    extraction_type: str

    report_folder: Path

    log_folder: Path

    seeker: FileSeekerBase

    jinja_environment = Environment

    processing_time: float

    input_path: Path

    output_folder = OutputFolder()

    def __init__(self, output_folder: Path, input_path: Path, device_type: str) -> None:
        self.default_configs = {
            "thumbnail_root": "**/Media/PhotoData/Thumbnails/**",
            "media_root": "**/Media",
            "thumbnail_size": (256, 256),
        }

        self.project = __project__
        self.version = __version__
        self.output_folder = output_folder
        self.create_output_folder()
        self.input_path = input_path
        self.device["type"] = device_type

    @cached_property
    def jinja_env(self) -> Environment:
        return self.create_jinja_environment()

    def create_output_folder(self):
        now = datetime.datetime.now()
        current_time = now.strftime("%Y-%m-%d_%A_%H%M%S")

        rf = self.report_folder = (
            Path(self.output_folder) / f"xLEAPP_Reports_{current_time}"
        )

        tf = self.temp_folder = rf / "temp"
        lf = self.log_folder = rf / "Script Logs"
        lf.mkdir(parents=True, exist_ok=True)
        tf.mkdir(parents=True, exist_ok=True)

    def create_jinja_environment(self) -> Environment:
        template_loader = jinja2.PackageLoader("xleapp.templating", "templates")
        log_file_loader = jinja2.FileSystemLoader(self.log_folder)

        options = {
            "loader": jinja2.ChoiceLoader([template_loader, log_file_loader]),
            "autoescape": jinja2.select_autoescape(["html", "xml"]),
            "extensions": [jinja2.ext.do, IncludeLogFileExtension],
            "trim_blocks": True,
            "lstrip_blocks": True,
        }
        rv = self.jinja_environment(**options)
        rv.globals.update(g=self)

        return rv

    @cached_property
    def artifacts(self) -> Enum:
        return Artifacts(self)

    def crunch_artifacts(self) -> None:
        return artifacts.crunch_artifacts(self)

    def generate_artifact_table(self):
        return artifacts.generate_artifact_table(self.artifacts)

    def generate_artifact_path_list(self):
        return artifacts.generate_artifact_path_list(self.artifacts)

    def generate_reports(self):
        logger_log.info("\nGenerating artifact report files...")
        nav = templating.generate_nav(
            report_folder=self.report_folder, artifacts=self.artifacts
        )
        for artifact in artifacts.filter_artifacts(self.artifacts.values()):
            msg_artifact = f"-> {artifact.category} [{artifact.cls_name}]"
            if not artifact.core:
                if artifact.report:
                    html_report = templating.ArtifactHtmlReport(
                        report_folder=self.report_folder,
                        log_folder=self.log_folder,
                        extraction_type=self.extraction_type,
                        navigation=nav,
                    )
                    if html_report(artifact).report:
                        logger_log.info(f"{msg_artifact}")
                else:
                    logger_log.warn(
                        f"{msg_artifact}: "
                        "Report not generated! Artifact "
                        "marked for no report generation. Check "
                        "artifact's 'report' attribute."
                    )
            else:
                logger_log.info(f"{msg_artifact}: Report skipped for core artifacts!")
        logger_log.info("Report files generated!")

    @property
    def num_to_process(self):
        return len(
            {artifact.cls_name for artifact in self.artifacts if artifact.selected}
        )

    @property
    def num_of_categories(self):
        return len(
            {
                artifact.value.category
                for artifact in self.artifacts
                if artifact.selected
            }
        )



import jinja2
from jinja2.ext import Extension


class IncludeLogFileExtension(Extension):
    tags = {"include_logfile"}

    def parse(self, parser):
        lineno = parser.stream.expect("name:include_logfile").lineno
        filename = parser.parse_expression()
        result = self.call_method("_render", [filename], lineno=lineno)
        return jinja2.nodes.Output([result], lineno=lineno)

    def _render(self, filename):
        try:
            file_text = "".join(
                (
                    self.environment.loader.get_source(self.environment, filename)[0]
                ).split("\r")
            )
        except:
            file_text = f"{filename} not found or missing! No logs available."
        return jinja2.Markup(f'<pre class="log-file">{file_text}</pre>')



import importlib
import inspect
import itertools
import logging
import typing as t

from dataclasses import dataclass
from enum import Enum
from importlib.metadata import entry_points
from pathlib import Path

from xleapp.helpers.strings import split_camel_case


if t.TYPE_CHECKING:
    from xleapp.app import XLEAPP

    from ._abstract import Artifact

logger_log = logging.getLogger("xleapp.logfile")


class Artifact:
    @property
    def cls_name(self) -> str:
        return self.value.cls_name.lower()

    @property
    def core(self) -> bool:
        return self.value.core

    @property
    def long_running_process(self) -> bool:
        return self.value.long_running_process

    @property
    def selected(self) -> bool:
        return self.value._selected

    @selected.setter
    def selected(self, bool: bool = True) -> None:
        """Toggles the selected state of the artifact

        Note:
            True fips the toggle state. i.e. `selected` property is changed from True
            to False or from False to True. False keeps the current state.

        Args:
            bool (bool, optional): Selected state of the artifact. Defaults to True.
        """
        if not self.core:
            self.value._selected ^= bool

    def process_artifact(self) -> None:
        msg_artifact = f"{self.value.category} [{self.cls_name}] artifact"
        logger_log.info(f"\n{msg_artifact} processing...")
        self.value.process_time, _ = self.value.process()
        if not self.value.processed:
            logger_log.warn(f"-> Artifact failed to processed!")
        logger_log.info(f"{msg_artifact} finished in {self.value.process_time:.2f}s")


class Artifacts:

    data: Enum

    def __init__(self, app: "XLEAPP") -> None:
        self.data = Artifacts.generate_artifact_enum(app)

    def __getattr__(self, name: str) -> t.Any:
        try:
            return getattr(self.data, name)
        except AttributeError:
            raise AttributeError(
                f"{__name__!r} doesn't have {name!r} attribute"
            ) from None

    @property
    def installed(self) -> list:
        return [artifact.cls_name for artifact in self]

    @property
    def selected_artifacts(self) -> list:
        return [artifact.cls_name for artifact in self.data if artifact.selected]

    def reset_selected(self):
        for artifact in self.data:
            if not artifact.core:
                artifact.selected = False

    def select_all(
        self,
        long_running_process: t.Optional[bool] = False,
    ) -> None:
        for artifact in self.data:
            if not artifact.core:
                if not long_running_process and artifact.long_running_process:
                    artifact.selected = False
                else:
                    artifact.selected = True

    @staticmethod
    def generate_artifact_enum(app: "XLEAPP"):
        members = {}
        device_type: str = app.device["type"]
        discovered_plugins = [
            plugin
            for plugin in entry_points()["xleapp.plugins"]
            if plugin.name == device_type
        ]

        for plugin in discovered_plugins:
            # Plugins return a str which is the plugin direction to
            # find plugins inside of. This direction is loading
            # that directory. Then, all the plugins are loaded.
            module_dir = Path(plugin.load()())

            for it in module_dir.glob("*.py"):
                if it.suffix == ".py" and it.stem not in ["__init__"]:
                    module_name = f'{".".join(module_dir.parts[-2:])}.{it.stem}'
                    module = importlib.import_module(module_name)
                    module_members = inspect.getmembers(module, inspect.isclass)
                    for name, xleapp_cls in module_members:
                        # check MRO (Method Resolution Order) for
                        # Artifact classes. Also, insure
                        # we do not get an abstract class.
                        if (
                            len(
                                {str(name).find("Artifact") for name in xleapp_cls.mro()}
                                - {-1},
                            )
                            != 0
                            and not inspect.isabstract(xleapp_cls)
                        ):
                            artifact: "Artifact" = dataclass(
                                xleapp_cls, unsafe_hash=True
                            )()
                            artifact_name = "_".join(
                                split_camel_case(artifact.cls_name)
                            ).upper()
                            artifact.app = app
                            members[artifact] = [
                                artifact_name,
                                artifact.cls_name.lower(),
                            ]

        return Enum(
            value=f"{device_type.title()}Artifact",
            names=itertools.chain.from_iterable(
                itertools.product(v, [k]) for k, v in members.items()
            ),
            module=__name__,
            type=Artifact,
        )



import functools
import logging
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from xleapp import artifacts

import xleapp.report as report
import xleapp.globals as g

if t.TYPE_CHECKING:
    from xleapp.artifacts._abstract import Artifact
    from xleapp.report import WebIcon

logger_log = logging.getLogger("xleapp.logfile")


class Template:
    """Template decorator for HTML pages

    This is used to decorate a custom class's HTML function to change
    the Jinja2 template being used.

    Example:

        Create a new class extending from HtmlPage base class. Then,
        create a `html()` function which is decorated by this function
        which provides the template name. This function will attached
        the `template` attribute automatically which is used to return the
        HTML of the template page.

        This replaces the default report page with this template.

        class MyCustomPage(HtmlPage):
            @Template('mytemplate')
            def html(self) -> str:
                return self.template.render(renderingVals)

     Args:
         template(str): Name of Jinja template without extension.
             Must be in template folder.
    """

    def __init__(self, template: str):
        self._template = f"{template}.jinja"

    def __call__(self, func):
        @functools.wraps(func)
        def template_wrapper(cls, *args):
            if not isinstance(cls, HtmlPage):
                TypeError(
                    f"{cls.__name__!r} not {HtmlPage.__name__!r} class for"
                    f" using {__name__!r} as a decorator!"
                )
            template_j = g.app.jinja_env.get_template(self._template)
            cls.template = template_j
            return func(cls)

        return template_wrapper


@dataclass
class _HtmlPageBase:
    device: object = field(init=False)


@dataclass
class _HtmlPageDefaults:
    """HTML page defaults for the HTML page

    Attributes:
        report_folder (Path): Report folder where Artifact HTML Reports
            are saved.
        log_folder (Path): Log folder for the log output
        extraction_type (str): Type of the extraction
        processing_time (float): Float number for time it took to run
            application
        device (Device): Extracted device information object
        project (str): project name
        version (str): project version
        navigation (dict): Navigation of the HTML report
    """

    report_folder: Path = field(default="", init=True)
    log_folder: Path = field(default="", init=True)
    extraction_type: str = field(default="fs", init=True)
    processing_time: float = field(default=0.0, init=True)
    navigation: dict = field(default_factory=lambda: {}, init=True)


@dataclass
class HtmlPage(ABC, _HtmlPageDefaults, _HtmlPageBase):
    def __call__(self, artifact: "Artifact"):
        self.artifact = artifact
        self.data = getattr(artifact, "data", None)
        return self

    @abstractmethod
    def html(self):
        raise NotImplementedError('HtmlPage objects must implement "html()" method!')


@dataclass
class Contributor:
    """Contributor's information displayed on main index page.

    Attributes:
        name (str): Name of Contributor
        website(str): Website of Contributor
        twitter (str): Twitter handle of the Contributor
        github (str): Github url for Contributor
    """

    name: str
    website: str = ""
    twitter: str = ""
    github: str = ""


@dataclass
class NavigationItem:
    """Navigation item for HTML report

    Attributes:
        name (str): Artifact Name
        href (str): URL of Artifact's report
        web_icon (webicons.Icon): Feather ICON for Artifact in
            the navigation list
    """

    name: str
    href: str
    web_icon: "WebIcon"

    def __str__(self) -> str:
        return f'<a class="nav-link" href="{self.href}"><span data-feather="{self.web_icon}"></span>{self.name}</a>'

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        return self.name == other.name


@dataclass
class ArtifactHtmlReport(HtmlPage):
    """Base Artifact HTML Report"""

    @Template("report_base")
    def html(self):
        """Returns HTML str of artifact report

        The :func:`@Template("report_base")` is the default Jinja2 template used for
        artifacts. Create a Jinja2 template in `report/templates` and then override
        the :func:`@html()`.

        Returns:
            str: HTML str of artifact report
        """
        return self.template.render(artifact=self.artifact, navigation=self.navigation)

    @property
    def report(self) -> bool:
        """Generates report information (html, tsv, kml, and timeline)"""
        html = self.html()
        output_file = (
            self.report_folder / f"{self.artifact.category} - {self.artifact.name}.html"
        )
        output_file.write_text(html)

        if self.artifact.processed and hasattr(self.artifact, "data"):
            options = (
                {
                    "name": self.artifact.name,
                    "data_list": self.data,
                    "data_headers": self.artifact.report_headers,
                },
            )
            report.save_to_db(
                report_folder=self.report_folder, db_type="tsv", options=options
            )

            if self.artifact.kml:
                report.save_to_db(
                    report_folder=self.report_folder,
                    db_type="kml",
                    options=options,
                )

            if self.artifact.timeline:
                report.save_to_db(
                    report_folder=self.report_folder,
                    db_type="timeline",
                    options=options,
                )
        return True

    @property
    def artifact_cls(self) -> str:
        return self.artifact.cls_name



from functools import cache, cached_property
import typing as t
from collections import defaultdict
from os import PathLike
from pathlib import Path

# from xleapp.artifacts.services import ArtifactService

from ._ext import IncludeLogFileExtension
from ._html import ArtifactHtmlReport, Contributor, HtmlPage, NavigationItem, Template
from ._partials.index import Index

if t.TYPE_CHECKING:
    from xleapp.app import XLEAPP


def generate_index(app: "XLEAPP") -> None:

    nav = generate_nav(app.report_folder, app.artifacts)

    index_page = Index(
        app.report_folder, app.log_folder, app.extraction_type, app.processing_time, nav
    )
    index_file = app.report_folder / "index.html"
    index_file.write_text(index_page.html())


def get_contributors(contributors: list) -> list[Contributor]:
    """Returns a list of Contributors from `xleapp.__contributors__`

    Args:
        contributors (list): List of contributors

    Returns:
        List(Contributor): List of contributors objects for HTML Report
    """
    contrib_list = []

    for row in contributors:
        contrib_list.append(Contributor(*row))
    return contrib_list


def generate_nav(report_folder: PathLike, artifacts) -> dict:
    """Generates a dictionary containing the navigation of the
       report.

    Args:
        report_folder (PathLike): Report folder where Artifact HTML Reports
            are saved.
        artifacts (ArtifactService): service containing all artifacts
            Artifacts for the report

    Returns:
        dict: dictionary of navigation items for HTML Report
    """
    nav = defaultdict(set)

    for artifact in artifacts:
        if not artifact.core and artifact.selected:
            temp_item = NavigationItem(
                name=artifact.value.name,
                web_icon=artifact.value.web_icon.value,
                href=(
                    report_folder / f"{artifact.value.category} - {artifact.name}.html"
                ),
            )
            nav[artifact.value.category].add(temp_item)
    return nav



import logging
import re
import webbrowser
from pathlib import Path

import PySimpleGUI as PySG

import xleapp.artifacts as artifacts
import xleapp.globals as g
from xleapp.__main__ import _main


class Handler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)
        self.setLevel(logging.INFO)

    def emit(self, record):
        global buffer
        record = f"{record.name}, [{record.levelname}], {record.message}"
        buffer = f"{buffer}\n{record}".strip()
        window["-LOG-"].update(value=buffer)


regex = re.compile(r"-Artifact(\w+)-")

artifact_list: dict = {}
num_of_artifacts: int = 0
buffer: str = ""
window: PySG.Window = None


def generate_layout(artifact_full_list: list) -> list:
    num_of_artifacts = len(artifact_list)
    mlist = [checklist(artifact) for artifact in sorted(artifact_full_list.items())]
    normal_font = ("Helvetica", 12)

    layout = [
        [PySG.Text("iOS Logs, Events, And Plists Parser", font=("Helvetica", 22))],
        [PySG.Text("https://github.com/abrignoni/xLEAPP", font=("Helvetica", 14))],
        [
            PySG.Frame(
                layout=[
                    [
                        PySG.Input(size=(97, 1)),
                        PySG.FileBrowse(
                            font=normal_font,
                            button_text="Browse File",
                            key="-INPUTFILEBROWSE-",
                        ),
                        PySG.FolderBrowse(
                            font=normal_font,
                            button_text="Browse Folder",
                            target=(PySG.ThisRow, -2),
                            key="-INPUTFOLDERBROWSE-",
                        ),
                    ],
                ],
                title=(
                    "Select the file type or directory of the"
                    "target iOS full file system extraction"
                    "for parsing:"
                ),
            ),
        ],
        [
            PySG.Frame(
                layout=[
                    [
                        PySG.Input(size=(112, 1)),
                        PySG.FolderBrowse(
                            font=normal_font,
                            button_text="Browse Folder",
                        ),
                    ],
                ],
                title="Select Output Folder:",
            ),
        ],
        [PySG.Text("Available Modules")],
        [PySG.Button("SELECT ALL"), PySG.Button("DESELECT ALL")],
        [
            PySG.Column(
                mlist, size=(300, 310), scrollable=True, vertical_scroll_only=True
            ),
            PySG.Output(size=(85, 20), key="-LOG-"),
        ],
        [
            PySG.ProgressBar(
                max_value=num_of_artifacts,
                orientation="h",
                size=(86, 7),
                key="-PROGRESSBAR-",
                bar_color=("Red", "White"),
            ),
        ],
        [
            PySG.Submit("Process", font=normal_font),
            PySG.Button("Close", font=normal_font),
        ],
    ]
    return layout


# initialize CheckBox control with module name
def checklist(artifact, disable=False):
    # items in the if are modules that take a long time to run.
    # Deselects them by default.

    name, obj = artifact
    mtxt = f"{obj.category} [{name}]"
    obj.selected = dstate = not obj.long_running_process
    disable = obj.core

    return [
        PySG.Checkbox(
            mtxt,
            default=dstate,
            key=f"-Artifact{name.upper()}-",
            metadata="{name}",
            disabled=disable,
            enable_events=True,
        ),
    ]


def set_progress_bar(value: int) -> None:
    """Sets the progress bar for the GUI

    Args:
        value (int): how much progress is on the bar
    """
    window["-PROGRESSBAR-"].update_bar(value)


def main(alist: dict) -> None:
    artifact_list = alist

    # sets theme for GUI
    PySG.theme("DarkAmber")

    window = PySG.Window(f"{g.__version__}", generate_layout(artifact_list))

    logging.getLogger(__name__).addHandler(Handler())

    while True:
        event, values = window.read()

        if event in (None, "Close"):  # if user closes window or clicks cancel
            break

        if regex.match(event):
            match = regex.match(event)
            name = match.group(1)
            if values[event]:
                artifacts.select(artifact_list, name.lower())

        elif event == "SELECT ALL":
            for name in list(artifact_list):
                window[f"-Artifact{name.upper()}-"].update(True)
                artifacts.select(artifact_list, name.lower())

        elif event == "DESELECT ALL":
            for name, obj in artifact_list.items():
                if not obj.core:
                    window[f"-Artifact{name.upper()}-"].update(False)
                    artifacts.select(artifact_list, name.lower())

        elif event == "Process":
            # check is selections made properly; if not we
            # will return to input form without exiting
            # app altogether

            for name in list(artifact_list):
                window[f"-Artifact{name.upper()}-"].update(disabled=True)
            input_path = values[0]
            output_folder = values[1]

            _main(artifact_list, input_path, output_folder)

            report_path = Path(g.report_folder / "index.html").resolve()

            locationmessage = f"Report name: {report_path}"
            PySG.Popup("Processing completed", locationmessage)
            webbrowser.open_new_tab(f"{report_path}")
            break

    window.close()



import argparse
import logging
import time

import xleapp.globals as g
import xleapp.log as log
import xleapp.report as report
import xleapp.templating as templating

from ._version import __project__, __version__
from .app import XLEAPP
from .artifacts import generate_artifact_path_list, generate_artifact_table
from .helpers.decorators import timed
from .helpers.search import search_providers
from .helpers.utils import ValidateInput, generate_program_header

logger_log = logging.getLogger("xleapp.logfile")


def get_parser():
    parser = argparse.ArgumentParser(
        prog=__project__.lower(),
        description="xLEAPP: iOS Logs, Events, and Plists Parser.",
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-I",
        dest="device_type",
        action="store_const",
        const="ios",
        help="parse ios artifacts",
    )
    group.add_argument(
        "-R",
        dest="device_type",
        action="store_const",
        const="returns",
        help="parse Warrant Returns / User Generated Archives artifacts",
    )
    group.add_argument(
        "-A",
        dest="device_type",
        action="store_const",
        const="android",
        help="parse android artifacts",
    )
    group.add_argument(
        "-C",
        dest="device_type",
        action="store_const",
        const="chromebook",
        help="parse Chromebook artifacts",
    )
    group.add_argument(
        "-V",
        dest="device_type",
        action="store_const",
        const="vehicle",
        help="parse vehicle artifacts",
    )
    parser.add_argument(
        "-o",
        "--output_folder",
        required=False,
        action="store",
        help="Output folder path",
    )
    parser.add_argument(
        "-i",
        "--input_path",
        required=False,
        action="store",
        help="Path to input file/folder",
    )
    parser.add_argument(
        "--artifact",
        required=False,
        action="store",
        help=(
            "Filtered list of artifacts to run. "
            "Allowed: core, <check artifact list in documentation>"
        ),
        metavar=None,
        nargs="*",
    )
    parser.add_argument(
        "-p",
        "--artifact_paths",
        required=False,
        action="store_true",
        help="Text file list of artifact paths",
    )
    parser.add_argument(
        "-l",
        "--artifact_table",
        required=False,
        action="store_true",
        help="Text file with table of artifacts",
    )
    group.add_argument(
        "--gui",
        required=False,
        action="store_true",
        help=f"Runs {__project__} into graphical mode",
    )
    parser.add_argument(
        "--version", action="version", version="f{__project__} v{__version__}"
    )

    return parser


def parse_args(parser):
    args = parser.parse_args()

    g.app = XLEAPP(
        output_folder=args.output_folder,
        input_path=args.input_path,
        device_type=args.device_type,
    )

    log.init()

    g.app.extraction_type = ValidateInput(
        args.input_path,
        args.output_folder,
        g.app.artifacts.selected,
    )

    # If an Itunes backup, use that artifact otherwise use
    # 'LastBuild' for everything else.
    if g.app.extraction_type == "itunes":
        g.app.artifacts.LAST_BUILD.value.selected = False
    else:
        g.app.artifacts.ITUNES_BACKUP_INFO.value.selected = False

    if args.artifact is None:
        # If no artifacts selected then choose all of them.
        g.app.artifacts.select_artifact(all_artifacts=True)
    else:
        filtered_artifacts = filter(
            lambda artifact: (artifact.lower() != 'core'), g.app.artifacts
        )
        for name in filtered_artifacts:
            try:
                g.app.artifacts[name.lower()].selected()
            except KeyError:
                g.app.error(
                    f"Artifact ({name.lower()}) not installed " f" or is unknown."
                )

    if args.gui:
        import xleapp.gui as gui

        gui.main(g.app)
    elif args.artifact_paths:
        generate_artifact_path_list()
    elif args.artifact_table:
        generate_artifact_table()
    else:
        return args
    exit()


def _main(app: "XLEAPP"):

    start_time = time.perf_counter()

    print(
        generate_program_header(
            f"{__project__} v{__version__}",
            app.input_path,
            app.report_folder,
            app.num_to_process,
            app.num_of_categories,
        ),
    )

    app.seeker = search_providers.create(
        app.extraction_type.upper(),
        directory=app.input_path,
        temp_folder=app.temp_folder,
    )

    @timed
    def process():
        logger_log.info(f"Processing {app.num_to_process} artifacts...")

        app.crunch_artifacts()

    run_time, _ = process()

    logger_log.info(f"\nCompleted processing artifacts in {run_time:.2f}s")
    report.copy_static_files(app.report_folder)

    end_time = time.perf_counter()

    app.processing_time = end_time - start_time

    logger_log.info("\nGenerating index file...")
    templating.generate_index(app)
    logger_log.info("-> Index file generated!")

    app.generate_reports()


def cli():
    parser = get_parser()
    args = parse_args(parser)

    _main(g.app)


if __name__ == "__main__":
    cli()



import typing as t
import re


def raw(data: t.ByteString) -> str:
    """Returns string of printable characters. Replacing non-printable characters
    with '.', or CHR(46)
    ``"""
    return "".join(
        [chr(byte) if byte >= 0x20 and byte < 0x7F else chr(46) for byte in data]
    )


def print(data: t.ByteString) -> filter:
    """Returns string of printable characters. Works similar to the Linux
    `string` function.
    """
    cleansed = "".join(
        [chr(byte) if byte >= 0x20 and byte < 0x7F else chr(0) for byte in data]
    )
    return filter(lambda string: len(string) >= 4, cleansed.split(chr(0)))


def split_camel_case(value: str) -> list[str]:
    return re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", value)).split()



import binascii
import math
from dataclasses import dataclass

import xleapp.helpers.strings as istrings
from xleapp.templating._html import HtmlPage, Template


@dataclass
class HexDumpHtml(HtmlPage):
    @Template("hexdump")
    def html(self, data, char_per_row=5):
        """Retuns HTML table of the hexdump of the passed in data."""
        data_hex = binascii.hexlify(data).decode("utf-8")
        str_raw = istrings.strings_raw(data)
        str_hex = ""
        str_ascii = ""

        """ Generates offset column
        """
        offset_rows = math.ceil(len(data_hex) / (char_per_row * 2))
        offsets = [i for i in range(0, len(data_hex), char_per_row)][:offset_rows]
        str_offset = "<br>".join([str(hex(s)[2:]).zfill(4).upper() for s in offsets])

        """ Generates hex data column
        """
        c = 0
        for i in range(0, len(data_hex), 2):
            str_hex += data_hex[i : i + 2] + "&nbsp;"

            if c == char_per_row - 1:
                str_hex += "<br>"
                c = 0
            else:
                c += 1

        """ Generates ascii column of data
        """
        for i in range(0, len(str_raw), char_per_row):
            str_ascii += str_raw[i : i + char_per_row] + "<br>"

        return self.template.render(str_offset, str_hex, str_ascii)



import importlib
import logging
import logging.config
import typing as t
from pathlib import Path
import os

import yaml

import xleapp.globals as g

from ..helpers.utils import generate_program_header

if t.TYPE_CHECKING:
    from xleapp.app import XLEAPP

StrPath = t.Union[str, os.PathLike[str]]


class ProcessFileFilter(logging.Filter):
    def filter(self, record):
        return record.name == "xleapp.process" and record.levelno >= 20


class InfoLogFileFilter(logging.Filter):
    def filter(self, record):
        return record.name == "xleapp.logfile" and record.levelno >= 20


class DebugFileFilter(logging.Filter):
    def filter(self, record):
        return g.app.debug


class StreamHandler(logging.StreamHandler):
    def emit(self, record: logging.LogRecord):
        if record.msg.startswith("->"):
            record.msg = f"    {record.msg}"
        logging.StreamHandler.emit(self, record)


class FileHandler(logging.FileHandler):
    def __init__(
        self,
        filename: StrPath,
        mode: str = "a",
        encoding: t.Union[str, None] = None,
        delay: bool = 0,
        errors: t.Union[str, None] = None,
    ) -> None:
        super().__init__(
            filename, mode=mode, encoding=encoding, delay=delay, errors=errors
        )

    def emit(self, record: logging.LogRecord):
        if record.msg.startswith("->"):
            record.msg = f"    {record.msg}"
        logging.FileHandler.emit(self, record)


class FileHandlerWithHeader(logging.FileHandler):
    def __init__(self, filename, header, mode="a", encoding=None, delay=0):
        self.header = header
        self.file_pre_exists = Path(filename)
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)
        if not delay and self.stream is not None:
            self.stream.write("%s\n" % header)

    def emit(self, record: logging.LogRecord):
        if self.stream is None:
            self.stream = self._open()
            if not self.file_pre_exists:
                self.stream.write("%s\n" % self.header)

        message = record.msg
        if message.startswith("->"):
            message = f"    {message}"
        record.msg = message
        logging.FileHandler.emit(self, record)


def init():
    logConfig = (
        Path(importlib.util.find_spec(__name__).origin).parent / "log_config.yaml"
    )
    with open(logConfig, "r") as file:
        config = yaml.safe_load(file.read())

    if not g.app.log_folder.exists():
        g.app.log_folder.mkdir(parents=True, exist_ok=True)

    info_log_file = config["handlers"]["info_file_handler"]["filename"]
    config["handlers"]["info_file_handler"]["filename"] = (
        g.app.log_folder / info_log_file
    )
    config["handlers"]["info_file_handler"]["header"] = generate_program_header(
        project_version=f"{g.app.project} v{g.app.version}",
        input_path=g.app.input_path,
        output_path=g.app.output_folder,
        num_to_process=g.app.num_to_process,
        num_of_categories=g.app.num_of_categories,
    )

    process_log_file = config["handlers"]["process_file_handler"]["filename"]
    config["handlers"]["process_file_handler"]["filename"] = (
        g.app.log_folder / process_log_file
    )

    debug_log_file = config["handlers"]["debug_file_handler"]["filename"]
    config["handlers"]["debug_file_handler"]["filename"] = (
        g.app.log_folder / debug_log_file
    )

    logging.config.dictConfig(config)



from dataclasses import dataclass, field

from xleapp._authors import __authors__, __contributors__

import xleapp.templating as templating
from .._html import HtmlPage, Template


@dataclass
class Index(HtmlPage):
    """Main index page for HTML report

    Attributes:
        authors (list): list of authors
        contributors (list): list of contributors
    """

    authors: list = field(init=False)
    contributors: list = field(init=False)

    def __post_init__(self):
        self.authors = templating.get_contributors(__authors__)
        self.contributors = templating.get_contributors(__contributors__)

    @Template("index")
    def html(self) -> str:
        """Generates html for page

        Returns:
            str: HTML of the index page
        """

        return self.template.render(
            navigation=self.navigation,
            authors=self.authors,
            contributors=self.contributors,
        )



import functools
import time
import typing as t

F = t.TypeVar('F', bound=t.Callable[..., t.Any])

# timer function
def timed(func: F) -> F:
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def timed_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        return run_time, value

    return timed_wrapper



"""Enum for Feather.js icons.

    286 Icons as of FeatherJS v4.28.0
    https://github.com/feathericons/feather/releases/tag/v4.28.0
"""
from enum import Enum

WebIcon: Enum = Enum(  # type: ignore  # https://github.com/python/typeshed/issues/2305
    "WebIcon",
    (
        (name.replace("-", "_").upper(), name)
        for name in [
            "activity",
            "airplay",
            "alert-circle",
            "alert-octagon",
            "alert-triangle",
            "align-center",
            "align-justify",
            "align-left",
            "align-right",
            "anchor",
            "aperture",
            "archive",
            "arrow-down-circle",
            "arrow-down-left",
            "arrow-down-right",
            "arrow-down",
            "arrow-left-circle",
            "arrow-left",
            "arrow-right-circle",
            "arrow-right",
            "arrow-up-circle",
            "arrow-up-left",
            "arrow-up-right",
            "arrow-up",
            "at-sign",
            "award",
            "bar-chart-2",
            "bar-chart",
            "battery-charging",
            "battery",
            "bell-off",
            "bell",
            "bluetooth",
            "bold",
            "book-open",
            "book",
            "bookmark",
            "box",
            "briefcase",
            "calendar",
            "camera-off",
            "camera",
            "cast",
            "check-circle",
            "check-square",
            "check",
            "chevron-down",
            "chevron-left",
            "chevron-right",
            "chevron-up",
            "chevrons-down",
            "chevrons-left",
            "chevrons-right",
            "chevrons-up",
            "chrome",
            "circle",
            "clipboard",
            "clock",
            "cloud-drizzle",
            "cloud-lightning",
            "cloud-off",
            "cloud-rain",
            "cloud-snow",
            "cloud",
            "code",
            "codepen",
            "codesandbox",
            "coffee",
            "columns",
            "command",
            "compass",
            "copy",
            "corner-down-left",
            "corner-down-right",
            "corner-left-down",
            "corner-left-up",
            "corner-right-down",
            "corner-right-up",
            "corner-up-left",
            "corner-up-right",
            "cpu",
            "credit-card",
            "crop",
            "crosshair",
            "database",
            "delete",
            "disc",
            "divide-circle",
            "divide-square",
            "divide",
            "dollar-sign",
            "download-cloud",
            "download",
            "dribbble",
            "droplet",
            "edit-2",
            "edit-3",
            "edit",
            "external-link",
            "eye-off",
            "eye",
            "facebook",
            "fast-forward",
            "feather",
            "figma",
            "file-minus",
            "file-plus",
            "file-text",
            "file",
            "film",
            "filter",
            "flag",
            "folder-minus",
            "folder-plus",
            "folder",
            "framer",
            "frown",
            "gift",
            "git-branch",
            "git-commit",
            "git-merge",
            "git-pull-request",
            "github",
            "gitlab",
            "globe",
            "grid",
            "hard-drive",
            "hash",
            "headphones",
            "heart",
            "help-circle",
            "hexagon",
            "home",
            "image",
            "inbox",
            "info",
            "instagram",
            "italic",
            "key",
            "layers",
            "layout",
            "life-buoy",
            "link-2",
            "link",
            "linkedin",
            "list",
            "loader",
            "lock",
            "log-in",
            "log-out",
            "mail",
            "map-pin",
            "map",
            "maximize-2",
            "maximize",
            "meh",
            "menu",
            "message-circle",
            "message-square",
            "mic-off",
            "mic",
            "minimize-2",
            "minimize",
            "minus-circle",
            "minus-square",
            "minus",
            "monitor",
            "moon",
            "more-horizontal",
            "more-vertical",
            "mouse-pointer",
            "move",
            "music",
            "navigation-2",
            "navigation",
            "octagon",
            "package",
            "paperclip",
            "pause-circle",
            "pause",
            "pen-tool",
            "percent",
            "phone-call",
            "phone-forwarded",
            "phone-incoming",
            "phone-missed",
            "phone-off",
            "phone-outgoing",
            "phone",
            "pie-chart",
            "play-circle",
            "play",
            "plus-circle",
            "plus-square",
            "plus",
            "pocket",
            "power",
            "printer",
            "radio",
            "refresh-ccw",
            "refresh-cw",
            "repeat",
            "rewind",
            "rotate-ccw",
            "rotate-cw",
            "rss",
            "save",
            "scissors",
            "search",
            "send",
            "server",
            "settings",
            "share-2",
            "share",
            "shield-off",
            "shield",
            "shopping-bag",
            "shopping-cart",
            "shuffle",
            "sidebar",
            "skip-back",
            "skip-forward",
            "slack",
            "slash",
            "sliders",
            "smartphone",
            "smile",
            "speaker",
            "square",
            "star",
            "stop-circle",
            "sun",
            "sunrise",
            "sunset",
            "tablet",
            "tag",
            "target",
            "terminal",
            "thermometer",
            "thumbs-down",
            "thumbs-up",
            "toggle-left",
            "toggle-right",
            "tool",
            "trash-2",
            "trash",
            "trello",
            "trending-down",
            "trending-up",
            "triangle",
            "truck",
            "tv",
            "twitch",
            "twitter",
            "type",
            "umbrella",
            "underline",
            "unlock",
            "upload-cloud",
            "upload",
            "user-check",
            "user-minus",
            "user-plus",
            "user-x",
            "user",
            "users",
            "video-off",
            "video",
            "voicemail",
            "volume-1",
            "volume-2",
            "volume-x",
            "volume",
            "watch",
            "wifi-off",
            "wifi",
            "wind",
            "x-circle",
            "x-octagon",
            "x-square",
            "x",
            "youtube",
            "zap-off",
            "zap",
            "zoom-in",
            "zoom-out",
        ]
    ),
)



__authors__ = [
    [
        "Alexis Brignoni",
        "https://abrignoni.com",
        "@AlexisBrignoni",
        "https://github.com/abrignoni",
    ],
    [
        "Yogesh Khatri",
        "https://swiftforensics.com",
        "@SwiftForensics",
        "https://github.com/ydkhatri",
    ],
]

__contributors__ = [
    ["Agam Dua", "https://loopback.dev", "@loopbackdev", "https://github.com/agamdua"],
    [
        "Sarah Edwards",
        "https://www.mac4n6.com",
        "@iamevltwin",
        "https://github.com/mac4n6",
    ],
    ["Heather Mahalik", "https://smarterforensics.com", "@HeatherMahalik", ""],
    ["Jessica Hyde", "https://twitter.com/B1N2H3X", "@B1N2H3X", ""],
    [
        "Phill Moore",
        "https://thinkdfir.com",
        "@phillmoore",
        "https://github.com/randomaccess3",
    ],
    ["Mattia Epifani", "https://blog.digital-forensics.it", "@mattiaep", ""],
    [
        "Mike Williamson",
        "https://forensicmike1.com",
        "@forensicmike1",
        "https://github.com/forensicmike",
    ],
    ["Geraldine Blay", "https://gforce4n6.blogspot.com", "@i_am_the_gia", ""],
    ["Christopher Vance", "https://blog.d204n6.com", "@cScottVance", ""],
    ["Brooke Gottlieb", "", "@xbrookego", ""],
    ["Jack Farley", "http://farleyforensics.com", "@JackFarley248", ""],
    ["Shafik Punja", "", "@qubytelogic", ""],
    [
        "Cheeky4N6Monkey",
        "https://cheeky4n6monkey.blogspot.com",
        "@Cheeky4n6Monkey",
        "https://github.com/cheeky4n6monkey",
    ],
    ["Edward Greybeard", "", "", "https://github.com/edward-greybeard"],
    ["Douglas Kein", "", "@DouglasKein", ""],
    ["Claudia Meda", "", "@KlodiaMaida", "https://github.com/KlodiaMaida"],
    ["Silvia Spallarossa", "", "@SilviaSpallaro1", "https://github.com/spatimbs"],
    ["Francesca Maestri", "", "@franc3sca_m", "https://github.com/francyM"],
    [
        "Christopher Vance",
        "https://blog.d204n6.com/",
        "@cScottVance",
        "https://github.com/cScottVance",
    ],
    [
        "Jesse Spangenberger",
        "https://cyberfenixtech.blogspot.com/",
        "@AzuleOnyx",
        "https://github.com/flamusdiu",
    ],
    [
        "Jon Baumann",
        "https://ciofecaforensics.com/",
        "@CiofecaForensic",
        "https://github.com/threeplanetssoftware",
    ],
    ["Scott Koenig", "", "@Scott_Kjr", ""],
    [
        "Kevin Pagano",
        "https://stark4n6.com/",
        "@KevinPagano3",
        "https://github.com/stark4n6",
    ],
    ["Ed Michael", "", "@EdXlg123", "https://github.com/edmichael"],
    ["Anna-Mariya Mateyna", "", "@any333_snickers", "https://github.com/any333"],
    ["Tommy Harris", "", "@tobraha", "https://github.com/tobraha"],
]



# -*- coding: utf-8 -*-
"""
Reports Module
"""

import importlib
import shutil
from pathlib import Path

from ._db import KmlDBManager, TimelineDBManager, TsvManager
from ._webicons import WebIcon


def save_to_db(report_folder: Path, db_type: str, **options) -> None:
    if db_type in ["kml", "timeline", "tsv"] and getattr(options, "data_list", []):
        if db_type == "kml":
            db = KmlDBManager(report_folder=report_folder, **options)
        if db_type == "tsv":
            db = TsvManager(report_folder=report_folder, **options)
        if db_type == "timeline":
            db = TimelineDBManager(report_folder=report_folder, **options)

        db.save()


def copy_static_files(report_folder: Path) -> None:
    """Copies static files to report folder.

    Returns:
        None
    """
    html_report_root = Path(importlib.util.find_spec(__name__).origin).parent
    static_folder = html_report_root / "_static"
    report_folder = report_folder / "_static"

    shutil.copytree(static_folder, report_folder, dirs_exist_ok=True)



import os
import sqlite3

from .utils import is_platform_windows


def open_sqlite_db_readonly(path: str) -> sqlite3.Connection:
    """Opens an sqlite db in read-only mode, so original db
    (and -wal/journal are intact)
    """
    if is_platform_windows():
        if path.startswith("\\\\?\\UNC\\"):  # UNC long path
            path = "%5C%5C%3F%5C" + path[4:]
        elif path.startswith("\\\\?\\"):  # normal long path
            path = "%5C%5C%3F%5C" + path[4:]
        elif path.startswith("\\\\"):  # UNC path
            path = "%5C%5C%3F%5C\\UNC" + path[1:]
        else:  # normal path
            path = "%5C%5C%3F%5C" + path
    return sqlite3.connect(f"file:{path}?mode=ro", uri=True)


def does_column_exist_in_db(
    db: sqlite3.Connection, table_name: str, col_name: str
) -> bool:
    """Checks if a specific col exists"""
    col_name = col_name.lower()
    try:
        db.row_factory = sqlite3.Row  # For fetching columns by name
        query = f"pragma table_info('{table_name}');"
        cursor = db.cursor()
        cursor.execute(query)
        all_rows = cursor.fetchall()
        for row in all_rows:
            if row["name"].lower() == col_name:
                return True
    except sqlite3.Error as ex:
        # logfunc(f"Query error, query={query} Error={str(ex)}")
        pass
    return False


def does_table_exist(db: sqlite3.Connection, table_name: str) -> bool:
    """Checks if a table with specified name exists in an sqlite db"""
    try:
        query = (
            f"SELECT name FROM sqlite_master "
            f"WHERE type='table' AND name='{table_name}'"
        )
        cursor = db.execute(query)
        for row in cursor:
            return True
    except sqlite3.Error as ex:
        # logfunc(f"Query error, query={query} Error={str(ex)}")
        pass
    return False



import os
import re
import typing as t
from pathlib import Path

from xleapp import __authors__


class ParseError(Exception):
    pass


def is_platform_windows() -> bool:
    """Returns True if running on Windows"""
    return os.name == "nt"


def sanitize_file_path(filename: str, replacement_char: str = "_") -> str:
    """
    Removes illegal characters (for windows) from the string passed.
    Does not replace \\ or /
    """
    return re.sub(r'[*?:"<>|\'\r\n]', replacement_char, filename)


def sanitize_file_name(filename: str, replacement_char: str = "_") -> str:
    """
    Removes illegal characters (for windows) from the string passed.
    """
    return re.sub(r'[\\/*?:"<>|\'\r\n]', replacement_char, filename)


def get_next_unused_name(path: str) -> str:
    """Checks if path exists, if it does, finds an unused name by appending -xx
    where xx=00-99. Return value is new path.
    If it is a file like abc.txt, then abc-01.txt will be the next
    """
    folder, basename = os.path.split(path)
    ext = None
    if basename.find(".") > 0:
        basename, ext = os.path.splitext(basename)
    num = 1
    new_name = basename
    if ext is None:
        new_name += f"{ext}"
    while os.path.exists(os.path.join(folder, new_name)):
        new_name = basename + "-{:02}".format(num)
        if ext is None:
            new_name += f"{ext}"
        num += 1
    return os.path.join(folder, new_name)


def ValidateInput(
    input_path: str,
    output_path: str,
    selected_artifacts: list,
) -> t.Union[str, str]:
    """
    Returns tuple (success, extraction_type)
    """

    if input_path is None:
        raise ParseError("No INPUT file or folder selected!")
    else:
        i_path = Path(input_path)  # input file/folder
        if not i_path.exists():
            raise ParseError("INPUT file/folder does not exist!")

    if output_path is None:
        raise ParseError("No OUTPUT folder selected!")
    else:
        o_path = Path(output_path)  # output folder
        if not o_path.exists():
            raise ParseError("OUTPUT does not exist!")

    if len(selected_artifacts) == 0:
        raise ParseError("No module selected for processing!")
    elif i_path.is_dir() and (i_path / "Manifest.db").exists():
        ext_type = "itunes"
    elif i_path.is_dir():
        ext_type = "fs"
    else:  # must be an existing file then
        ext_type = i_path.suffix[1:].lower()
        if ext_type not in ["tar", "zip", "gz"]:
            raise ParseError(f"Input file is not a supported archive! \n {i_path}")

    return ext_type


def generate_program_header(
    project_version: str,
    input_path: Path,
    output_path: Path,
    num_to_process: int,
    num_of_categories: int,
):
    project, version = project_version.split(" ")
    header = (
        "Procesing started. Please wait. This may take a "
        "few minutes...\n"
        "-----------------------------------------------------------"
        "---------------------------\n\n"
        f"{project} {version}: xLEAPP Logs, "
        "Events, and Properties Parser\n"
        "Objective: Triage iOS Full System Extractions.\n\n"
    )

    for author in __authors__:
        header += f"By: {author[0]} | {author[2]} | {author[1]}\n"

    header += (
        f"\nArtifacts to parse: {num_to_process} in {num_of_categories} "
        "categories\n"
        f"File/Path Selected: {input_path}\n"
        f"Output: {output_path}\n\n"
        "-----------------------------------------------------------"
        "---------------------------\n"
    )
    return header


def filter_dict(d: dict, filter_string: str):
    for key, val in d.items():
        if filter_string not in key:
            continue
        yield key, val



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['xleapp',
 'xleapp.artifacts',
 'xleapp.helpers',
 'xleapp.log',
 'xleapp.report',
 'xleapp.templating',
 'xleapp.templating._partials']

package_data = \
{'': ['*'],
 'xleapp.report': ['_static/*',
                   '_static/MDB-Free_4.13.0/css/*',
                   '_static/MDB-Free_4.13.0/css/addons/*',
                   '_static/MDB-Free_4.13.0/img/overlays/*',
                   '_static/MDB-Free_4.13.0/img/svg/*',
                   '_static/MDB-Free_4.13.0/js/*',
                   '_static/MDB-Free_4.13.0/js/addons/*',
                   '_static/MDB-Free_4.13.0/js/modules/*'],
 'xleapp.templating': ['templates/*']}

install_requires = \
['Jinja2>=3.0.2,<4.0.0',
 'PySimpleGUI>=4.49.0,<5.0.0',
 'PyYAML>=6.0,<7.0',
 'prettytable>=2.2.1,<3.0.0',
 'python-magic-bin>=0.4.14,<0.5.0',
 'simplekml>=1.3.6,<2.0.0']

entry_points = \
{'console_scripts': ['xleapp = xleapp._main__:cli']}

setup_kwargs = {
    'name': 'xleapp',
    'version': '0.1.0a0',
    'description': 'Multiplaform Logs, Events, And Plists Parser',
    'long_description': "# xLEAPP\n\n**Development build. Please be cauious using on real cases.**\n\nFramework for Logs, Events, And Plists Parser (LEAPP)\n\nThis framework is a complete rewrite of the excellent tool iLEAPP.Details of iLEAPP can be found in this [blog post](https://abrignoni.blogspot.com/2019/12/xleapp-ios-logs-events-and-properties.html)\n\nxLEAPP is the framework created to merge several tools together. More information about the rewrite is given in by talk ([YouTube](https://www.youtube.com/watch?v=seTpCmSF0Gc)) at Black Hills Info Security&#39;s Wild West Hackin&#39; Fest (WWHF): Deadwood in 2021.\n\n## Features\n\n* Provides a centralized and modular framework\n* Provides a simplified way to write plugins (artifacts) for each different supported platform.\n* Parses iOS, macOS, Android, Chromebook, warranty returns, and Windows artifacts depending on the plugins installed.\n\n## Other Documentation\n\n* [Artifact Creation](docs/current/artifact-creation.md)\n\n## Pre-requisites\n\nThis project requires you to have Python >= 3.9\n\n## Installation\n\n### Windows\n\n```powershell\nPS> py -m pip install xleapp\nPS> py -m pip install xleapp-<plugin>\n```\n\n### Linux\n\n```bash\n$ python -m pip install xleapp\n$ python -m pip install xleapp-<plugin>\n```\n\n## Installation from Github and Development Information\n\n* [Windows](docs/current/windows.md)\n* [Linux](docs/current/linux.md)\n\n## VS Code configuration files\n\nThere are several [configuration files](https://github.com/flamusdiu/xleapp-project) that I have been using for VS Code.\n\n## Compile to executable\n\n**NOTE:** This may not work at this time with this alpha version.\n\nTo compile to an executable so you can run this on a system without python installed.\n\nTo create xleapp.exe, run:\n\n```bash\npyinstaller --onefile xleapp.spec\n```\n\nTo create xleappGUI.exe, run:\n\n```bash\npyinstaller --onefile --noconsole xleappGUI.spec\n```\n\n## Usage\n\n### CLI\n\n```bash\n$ xleapp -h\nusage: xleapp [-h] [-I] [-R] [-A] [-C] [-V] [-o OUTPUT_FOLDER] [-i INPUT_PATH]\n       [--artifact [ARTIFACT ...]] [-p] [-l] [--gui] [--version]\n\nxLEAPP: Logs, Events, and Plists Parser.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -I                    parse ios artifacts\n  -R                    parse Warrant Returns / User Generated Archives artifacts\n  -A                    parse android artifacts\n  -C                    parse Chromebook artifacts\n  -V                    parse vehicle artifacts\n  -o OUTPUT_FOLDER, --output_folder OUTPUT_FOLDER\n                        Output folder path\n  -i INPUT_PATH, --input_path INPUT_PATH\n                        Path to input file/folder\n  --artifact [ARTIFACT ...]\n                        Filtered list of artifacts to run. Allowed: core, <check artifact list in\n                        documentation>\n  -p, --artifact_paths  Text file list of artifact paths\n  -l, --artifact_table  Text file with table of artifacts\n  --gui                 Runs xLEAPP into graphical mode\n  --version             show program&#39;s version number and exit\n\n```\n\n### GUI\n\nThis needs work and may not work properly!\n\n```bash\n$ xleapp --gui \n\n```\n\n### Help\n\n```bash\n$ xleapp.py --help\n\n```\n\nThe GUI will open in another window.  \n\n## Acknowledgements\n\nThis tool is the result of a collaborative effort of many people in the DFIR community.\n\nThis product includes software developed by Sarah Edwards (Station X Labs, LLC, @iamevltwin, mac4n6.com) and other contributors as part of APOLLO (Apple Pattern of Life Lazy Output'er).\n",
    'author': 'Jesse Spangenberger',
    'author_email': 'azuleonyx@digitalforensics.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/flamusdiu/xleapp',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import fnmatch
import logging
import os
import sqlite3
import tarfile
import typing as t
from abc import ABC, abstractmethod
from collections import UserDict
from functools import lru_cache
from io import IOBase
from pathlib import Path
from shutil import copyfile
from zipfile import ZipFile

from xleapp.helpers.descriptors import Validator

from .db import open_sqlite_db_readonly
from .utils import is_platform_windows

logger_log = logging.getLogger("xleapp.logfile")
logger_process = logging.getLogger("xleapp.process")


class PathValidator(Validator):
    def validator(self, value):
        if not isinstance(value, (Path, os.PathLike, str)):
            raise TypeError(f"Expected {value!r} to be a Path or Pathlike object")
        return value.resolve()


class HandleValidator(Validator):
    def validator(self, value):
        if isinstance(value, (str, Path)):
            return None
        elif not isinstance(value, (sqlite3.Connection, IOBase)):
            raise TypeError(
                f"Expected {value!r} to be one of: string, Path, sqlite3.Connection or IOBase."
            )


class Handle:
    h = HandleValidator()
    path = PathValidator()

    def __init__(self, file: t.Any, path: t.Any = None) -> None:
        self.path = path
        self.h = file

    def __call__(self):
        return self.h or self.path


class FileHandles(UserDict):
    logged: list = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)
        self.default_factory = set

    def __len__(self):
        return sum(count for count in self.values())

    def add(self, regex: str, files: list, file_names_only: bool = False) -> None:

        if self.get(regex):
            return

        for item in files:

            """
            If we have more then 10 files, then set only
            file names instead of `FileIO` or
            `sqlite3.connection` to save memory. Most artifacts
            probably have less then 5 files they will read/use.
            """
            h = None
            if len(files) > 10 or file_names_only:
                h = Handle(item)
            else:
                if item.drive.startswith("\\\\?\\"):
                    extended_path = Path(item)

                path = Path(item.resolve())

                try:
                    db = sqlite3.connect(f"file:{path}?mode=ro", uri=True)
                    cursor = db.cursor()
                    (page_size,) = cursor.execute("PRAGMA page_size").fetchone()
                    (page_count,) = cursor.execute("PRAGMA page_count").fetchone()

                    if page_size * page_count < 40960:  # less then 40 MB
                        db_mem = sqlite3.connect(":memory:")
                        db.backup(db_mem)
                        db.close()
                        db = db_mem
                    db.row_factory = sqlite3.Row
                    h = Handle(file=db, path=path)
                except (sqlite3.OperationalError, sqlite3.DatabaseError, TypeError):
                    fp = open(extended_path, "rb")
                    h = Handle(file=fp, path=path)
                except FileNotFoundError:
                    raise FileNotFoundError(f"File {path!r} was not found!")
            if h:
                self[regex].add(h)

    def __getitem__(self, regex: str) -> set[Handle]:
        try:
            files = super().__getitem__(regex)

            for num, file in enumerate(files, start=1):
                if regex not in self.logged:
                    if num == 1:
                        logger_process.info(f"\nFiles for {regex} located at:")

                    logger_process.info(f"    {file.path}")
                    self.logged.append(regex)

                if isinstance(file, IOBase):
                    file.seek(0)
            return files
        except KeyError:
            raise KeyError(f"Regex {regex} has no files opened!")

    def __delitem__(self, regex: str) -> None:
        files = self.__dict__.pop(regex, None)
        if files:
            if isinstance(files, list):
                for f in files:
                    f.close()
            else:
                files.close()

    def __missing__(self, key: str) -> set[Handle]:
        if self.default_factory is None:
            raise KeyError(key)
        if key not in self:
            self[key] = self.default_factory()
        return self[key]


class FileSeekerBase(ABC):

    temp_folder: Path
    _all_files: list[str] = []
    _directory: Path
    _file_handles = FileHandles()

    @abstractmethod
    def search(
        self, filepattern_to_search: str, return_on_first_hit: bool = False
    ) -> t.Iterator[t.Any]:
        """Returns a list of paths for files/folders that matched"""
        pass

    @abstractmethod
    def cleanup(self) -> None:
        """close any open handles"""
        pass

    @abstractmethod
    def build_files_list(
        self, folder: t.Optional[t.Union[str, Path]]
    ) -> t.Union[tuple[list, list], list]:
        """Finds files in directory"""
        pass

    @property
    def directory(self) -> Path:
        return self._directory

    @directory.setter
    def directory(self, directory: t.Union[str, Path]):
        self._directory = Path(directory)

    @property
    def all_files(self) -> list[str]:
        return self._all_files

    @all_files.setter
    def all_files(self, files: list[str]):
        self._all_files = files

    @property
    def file_handles(self) -> FileHandles:
        return self._file_handles


class FileSeekerDir(FileSeekerBase):
    def __init__(self, directory, temp_folder=None):
        self.directory = Path(directory)
        logger_log.info("Building files listing...")
        subfolders, files = self.build_files_list(directory)
        self.all_files.extend(subfolders)
        self.all_files.extend(files)
        logger_log.info(f"File listing complete - {len(self._all_files)} files")

    def build_files_list(self, folder):
        """Populates all paths in directory into _all_files"""

        subfolders, files = [], []

        for item in os.scandir(folder):
            if item.is_dir():
                subfolders.append(item.path)
            if item.is_file():
                files.append(item.path)

        for folder in list(subfolders):
            sf, items = self.build_files_list(folder)
            subfolders.extend(sf)
            files.extend(items)

        return subfolders, files

    def search(self, filepattern, return_on_first_hit=False):
        return iter(fnmatch.filter(self.all_files, filepattern))

    def cleanup(self):
        pass


class FileSeekerItunes(FileSeekerBase):
    def __init__(self, directory, temp_folder) -> None:

        self.directory = Path(directory)
        self.temp_folder = Path(temp_folder)

        self.build_files_list()

    def build_files_list(self, folder=None) -> list:
        """Populates paths from Manifest.db files into _all_files"""

        directory = self.directory

        db = open_sqlite_db_readonly(directory / "Manifest.db")
        cursor = db.cursor()
        cursor.execute(
            """
            SELECT
            fileID,
            relativePath
            FROM
            Files
            WHERE
            flags=1
            """,
        )
        all_rows = cursor.fetchall()
        for row in all_rows:
            hash_filename = row[0]
            relative_path = row[1]
            self._all_files[relative_path] = hash_filename
        db.close()

        # does not return any values for this FileSeeker
        return []

    def search(self, filepattern, return_on_first_hit=False):
        pathlist = []
        matching_keys = fnmatch.filter(self._all_files, filepattern)
        for relative_path in matching_keys:
            hash_filename = self._all_files[relative_path]
            original_location = Path(self.directory) / hash_filename[:2] / hash_filename
            temp_location = Path(self.temp_folder) / relative_path

            temp_location.mkdir(parents=True, exist_ok=True)
            copyfile(original_location, temp_location)
            pathlist.append(temp_location)

        return iter(pathlist)


class FileSeekerTar(FileSeekerBase):
    def __init__(self, directory, temp_folder):
        self.is_gzip = Path(directory).suffix == ".gz"
        mode = "r:gz" if self.is_gzip else "r"
        self.tar_file = tarfile.open(directory, mode)
        self.temp_folder = Path(temp_folder)
        self.tar_file.getmembers()

    def search(self, filepattern, return_on_first_hit=False):
        for member in self.build_files_list():
            if fnmatch.fnmatch(member.name, filepattern):

                if is_platform_windows():
                    full_path = Path(f"\\\\?\\{self.temp_folder / member.name}")
                else:
                    full_path = self.temp_folder / member.name

                if not member.isdir():
                    full_path.parent.mkdir(parents=True, exist_ok=True)
                    full_path.write_bytes(
                        tarfile.ExFileObject(self.tar_file, member).read(),
                    )
                    os.utime(full_path, (member.mtime, member.mtime))
                yield full_path

    def cleanup(self):
        self.tar_file.close()

    @lru_cache(maxsize=5)
    def build_files_list(self) -> list[tarfile.TarInfo]:
        return self.tar_file.getmembers()


class FileSeekerZip(FileSeekerBase):
    def __init__(self, zip_file_path, temp_folder):
        self.zip_file = ZipFile(zip_file_path)
        self.name_list = self.zip_file.namelist()
        self.temp_folder = temp_folder

    def search(self, filepattern, return_on_first_hit=False):
        pathlist = []
        for member in self.name_list:
            if fnmatch.fnmatch(member, filepattern):
                try:
                    extracted_path = (
                        # already replaces illegal chars with _ when exporting
                        self.zip_file.extract(member, path=self.temp_folder)
                    )
                    pathlist.append(extracted_path)
                except Exception:
                    member = member.lstrip("/")
                    # logfunc(f'Could not write file to filesystem, path was {member} ' + str(ex))
        return iter(pathlist)

    def cleanup(self):
        self.zip_file.close()


class FileSearchProvider(UserDict):
    def __init__(self) -> None:
        self.data = {}
        self._items = 0

    def __len__(self) -> int:
        return self._items

    def register_builder(self, key, builder) -> None:
        self._items = self._items + 1
        self.data[key] = builder

    def create(self, key, **kwargs) -> FileSeekerBase:
        builder = self.data.get(key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)


search_providers = FileSearchProvider()
search_providers.register_builder("FS", FileSeekerDir)
search_providers.register_builder("ITUNES", FileSeekerItunes)
search_providers.register_builder("TAR", FileSeekerTar)
search_providers.register_builder("ZIP", FileSeekerZip)



# coding: utf-8
import os
import json

"""
This helper renders chat conversations passed as a json dump of a dictionary:
    --conversation contact id/name
        --num message (start from 0 each time is fine)
            --data-name = correspondant (phone or ID)
            --data-time = time of message (formatted as str)
            --from_me = (boolean - 0 = received / 1 = sent)
            --message = message content

example:
    {
        "Vincent":{
            "0":{
                "data-name": "Vincent",
                "data-time": "2020-11-10 08:00:00",
                "from_me" : 0,
                "Message": "What is your favorite tool?"
            },
            "1":{
                "data-name": "Vincent",
                "data-time": "2020-11-10 08:01:00",
                "from_me" : 1,
                "Message": "xLEAPP !"
            },
        },
        "Mike,Vincent":{
            "0":{
                "data-name": "Mike",
                "data-time": "2020-11-10 08:08:00",
                "from_me" : 0,
                "Message": "Who like apples ?"
            },
            "1":{
                "data-name": "Vincent",
                "data-time": "2020-11-10 08:09:00",
                "from_me" : 1,
                "Message": "I do!"
            }
    }

"""

mimeTypeIcon = {
    "image": "📷",
    "audio": "🎧",
    "video": "🎥",
    "animated": "🎡",
    "application": "📎",
    "text": "Ŧ",
}


"""
helper to render body with attachments
"""


def integrateAtt(rec):
    if rec["file-path"]:
        att_type = (
            rec["content-type"].split("/")[0] if rec["content-type"] else "application"
        )
        filename = os.path.basename(rec["file-path"])
        body = rec["message"] if rec["message"] else ""
        if att_type == "image":
            source = '<img src="{}" width="256" height="256"/>'.format(rec["file-path"])

        elif att_type == "audio":
            source = """
            <audio controls>
              <source src="{0}" type="{1}">
              <p><a href="{0}"></a> </p>
            </audio>
            """.format(
                rec["file-path"], rec["content-type"]
            )
        elif att_type == "video":
            source = """
            <video controls width="256">
              <source src="{0}" type="{1}">
              <p><a href="{0}"></a> </p>
            </video>
            """.format(
                rec["file-path"], rec["content-type"]
            )
        else:
            source = '<a href="{}">{}</a>'.format(rec["file-path"], filename)

        return "\n".join([body, mimeTypeIcon[att_type] + " " + source])
    else:
        return rec["message"]


"""
transform a chat df to be rendered to js
input : df with following columns:
    - data-name str : contact name / number
    - data-time dt : time of message (needs to be datetime format)
    - message str : text message
    - content-type str : mime type of atachement or None (ex : 'image/jpeg')
    - file-path str : path of attachment to render
    - from_me bool : 0 if received, 1 if sent
output : 
    str including script and data to include in report html

"""


def render_chat(df):
    df["body_to_render"] = df.apply(lambda rec: integrateAtt(rec), axis=1)
    latest_mess = df.groupby("data-name", as_index=False)["data-time"].max()
    df = df.merge(
        latest_mess, on=["data-name"], how="right", suffixes=["", "_latest"]
    ).sort_values(by=["data-time_latest", "data-name"], ascending=[False, True])
    df["data-time"] = df["data-time"].dt.strftime("%Y-%m-%d %H:%M:%S")
    chats = {}
    for c in df["data-name"].unique():
        chats[c] = (
            df[df["data-name"] == c][
                ["data-name", "from_me", "body_to_render", "data-time"]
            ]
            .reset_index(drop=True)
            .to_dict(orient="index")
        )

    json_chat = json.dumps(chats)
    return render_js_chat(json_chat)



import os
import shutil

from PIL import (  # type: ignore  # https://github.com/python-pillow/Pillow/issues/2625
    Image,
)


def generate_thumbnail(imDirectory, imFilename, seeker, report_folder):
    """
    searching for thumbnails, copy it to report folder and return tag  to
    insert in html
    """
    thumb = f"{g.app.get('thumbnail_root')}/{imDirectory}/{imFilename}"
    thumblist = seeker.search(f"{thumb}/**.JPG", return_on_first_hit=True)
    thumbname = imDirectory.replace("/", "_") + "_" + imFilename + ".JPG"
    pathToThumb = os.path.join(
        os.path.basename(os.path.abspath(report_folder)), thumbname
    )
    htmlThumbTag = '<img src="{0}"></img>'.format(pathToThumb)
    if thumblist:
        shutil.copyfile(thumblist[0], os.path.join(report_folder, thumbname))
    else:
        # recreate thumbnail from image
        # TODO: handle videos and HEIC
        files = seeker.search(
            f"{g.app.get('MEDIA_ROOT')}/{imDirectory}/{imFilename}",
            return_on_first_hit=True,
        )
        if files:
            try:
                im = Image.open(files[0])
                im.thumbnail(g.app.get("THUMB_SIZE"))
                im.save(os.path.join(report_folder, thumbname))
            except:  # noqa
                pass  # unsupported format
    return htmlThumbTag



# -*- coding: utf-8 -*-
"""This module contains all the functions for Artifact abstract class.

Python requires that attributes with out default values come after attributes with defaults.
The class order ABC > _AbstractArtifactDefaults > _AbstractBase ensures this is correct.

:obj:`_AbstractBase` contains all attributes without a default. Please place any extra attributes
that do not require defaults within this class. Also ensure to use `field(init=false)` for each 
attribute so they are not required when the class is first created.

:obj:`_AbstractArtifactDefaults` provides any attributes with defaults. Also, use `field(init=False)`
for each one as before.

:obj:`Artifact` is the class every artifact needs to be subclassed from.

"""

import logging
import typing as t
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path

import xleapp.artifacts as artifacts

from ._descriptors import FoundFiles, ReportHeaders, WebIcon

if t.TYPE_CHECKING:
    from xleapp.app import XLEAPP


@dataclass
class _AbstractBase:
    """Base class to set any properties for
    `Artifact` Class. This properties do not
    have a default value.

    Attributes:
        category (str): Category the artifact falls into. This also is used for the area
            the artifacts appears under for the end report.
        description (str): Short description of the artifact.
        name (str): full name of the artifact.
        data (list): list of data from the `process()`.
        regex (list): search strings set by the `@Search()` decorator.
        _log (logging.Logger). Logger attached to artifact for output to log files.
    """

    description: str = field(init=False, repr=False, compare=False, hash=False)
    name: str = field(init=False, compare=True, hash=True)
    data: list = field(init=False, repr=False, compare=False, hash=False)
    regex: set = field(init=False, repr=False, compare=False, hash=False)
    app: "XLEAPP" = field(init=False, repr=False, compare=False, hash=False)
    _log: logging.Logger = field(init=False, repr=False, compare=False, hash=False)


@dataclass
class _AbstractArtifactDefaults:
    """Class to set defaults to any properties for the
    :obj:`Artifact` class.

    Attributes core, long_running_process, and selected are used
    to track artifacts internally for certain actions.

    Args:
        app (XLEAPP): attached app instance to each artifact. Default is None.
        core (bool): artifacts require to always run. Default is False.
        found (set): set of files found from a `FileSeeker`.
        hide_html_report_path_table (bool): bool to hide displaying paths
            when then report is generated. Note: Any artifact processing 10
            or more files will ignore this value.
        html_report (ArtifactHtmlReport): holds the html report object
        kml (bool): True or False to generate kml (location files) information.
            Default is False.
        long_running_process (bool): artifacts which takes an extremely long time to run
            and should be deselected by default. Default is False
        processing_time(float): Seconds of time it takes to process this artifact.
            Default is 0.0.
        report (bool): True or False. sets to generate HTML report. Default True.
        report_headers (list or tuple): headers for the report table during
            report generation.
        selected: artifacts selected to be run. Default is False.
        timeline(bool): True or False to add data to the timeline database. Default is False.
        web_icon (Icon): FeatherJS icon used for the report navgation menu. Default is `Icon.TRIANGLE`.
    """

    category: str = field(init=False, default="Unknown", hash=True, compare=True)
    core: bool = field(init=False, default=False)
    found: FoundFiles = field(init=False, default=FoundFiles(), hash=False)
    kml: bool = field(init=False, default=False)
    long_running_process: bool = field(init=False, default=False)
    processed: bool = field(init=False, default=False)
    process_time: float = field(init=False, default=float())
    report: bool = field(init=False, default=True)
    report_headers: ReportHeaders = field(
        init=False, default=ReportHeaders(), hash=False
    )
    _selected: bool = field(init=False, default=False)
    timeline: bool = field(init=False, default=False)
    web_icon: WebIcon = field(init=False, default=WebIcon())


@dataclass  # type: ignore  # https://github.com/python/mypy/issues/5374
class Artifact(ABC, _AbstractArtifactDefaults, _AbstractBase):
    """Abstract class for creating Artifacts"""

    @abstractmethod
    def process(self) -> None:
        """Gets artifacts located in `files_found` params by the
        `seeker`. It saves should save the report in `report_folder`.
        """
        raise NotImplementedError(
            "Needs to implement AbastractArtifact's" "process() method!",
        )

    @contextmanager
    def context(
        self,
        regex: set[str],
        file_names_only: bool = False,
        return_on_first_hit: bool = True,
    ) -> t.Iterator["Artifact"]:
        """Creates a contaxt manager for an artifact.

        This will automatically search and add the regex and files to an artifact when
        called. You then can use `self.found` when processing the artifact.

        Example:
            This can be used with `with` blocks::

                with Artifact.context(['*/myregex*'], file_names_only=True) as artifact:
                    artifact.name = 'New Name'
                    artifact.process()


        Args:
            regex (list[str]): Globs regex expressions
            file_names_only (bool, optional): True or False. Returns a list of file
                names (true) or list of objects (false) . Defaults to False.
            return_on_first_hit (bool, optional): True or False. Returns first match on
                a regex search. Defaults to True.

        Yields:
            Iterator[Artifact]: returns self
        """
        seeker = self.app.seeker
        files = seeker.file_handles
        global_regex = files.keys()

        self.regex = regex

        for r in self.regex:
            results = set()

            if r in global_regex:
                results = files[r]
            else:
                try:
                    if return_on_first_hit:
                        results = {next(seeker.search(r))}
                    else:
                        results = set(seeker.search(r))
                except StopIteration:
                    results = set()

                if bool(results):
                    files.add(r, results, file_names_only)

            if bool(results):
                if return_on_first_hit or len(results) == 1:
                    self.found = {files[r].copy().pop()}
                else:
                    self.found = self.found | files[r]
        yield self

    def __enter__(self):
        return self

    @property
    def cls_name(self) -> str:
        """Returns class Name of object

        Returns:
            str: class name
        """
        return type(self).__name__

    def copyfile(self, input_file: Path, output_file: str):
        return artifacts.copyfile(
            report_folder=self.app.report_folder,
            name=self.cls_name,
            input_file=input_file,
            output_file=output_file,
        )

    def log(
        self,
        level: int = logging.INFO,
        message: object = None,
    ):
        if not hasattr(self, "_log"):
            self._log = logging.getLogger("xleapp.logfile")

        if not message:
            raise AttributeError(f"Message missing for log on {self.cls_name}!")

        self._log.log(level, msg=message)



import xleapp.report._webicons as web
from xleapp.helpers.descriptors import Validator


class FoundFiles(Validator):
    """Descriptor ensuring 'foundfiles' type"""

    default_value = set()

    def validator(self, value):
        if not isinstance(value, set):
            raise TypeError(f"Expected {value!r} to be a set!")


class WebIcon(Validator):
    """Descriptor ensuring 'web_icon' type"""

    default_value = web.WebIcon.ALERT_TRIANGLE

    def validator(self, value):
        if not isinstance(value, (web.WebIcon, WebIcon)):
            raise TypeError(f"Expected {value!r} to be {web.WebIcon!r}! ")


class ReportHeaders(Validator):
    """Descriptor ensuring 'report_headers' type"""

    default_value = ("Key", "Value")

    def validator(self, value) -> None:
        if not ReportHeaders._check_list_of_tuples(value, bool_list=[]):
            raise TypeError(
                f"Expected {value!r} to be a list of tuples or tuple!",
            )

    @staticmethod
    def _check_list_of_tuples(values: list, bool_list: list[bool] = None) -> bool:
        """Checks list to see if its a list of tuples of strings

        Examples:

            Set headers for a single table
            self.report_headers = ('Timestamp', 'Account Desc.', 'Username',
                                   'Description', 'Identifier', 'Bundle ID')

            Set headers for more the one table
            self.report_headers = [('Timestamp', 'Account Desc.', 'Username',
                                    'Description', 'Identifier', 'Bundle ID'),
                                   ('Key', 'Value)]

            Anything else should fail to set.

        Args:
            values (list): values to be checked
            bool_list (list, optional): list of booleans of values checked.
                Defaults to [].

        Returns:
            bool: Returns true or false if values match for tuples of strings
        """
        bool_list = bool_list or []
        if values == []:
            return all(bool_list)
        else:
            if isinstance(values, tuple):
                idx = values
                values = []
            elif isinstance(values, list):
                idx = values[:1][0]
                values = values[1:]
            else:
                return False

            if isinstance(idx, list):
                return ReportHeaders._check_list_of_tuples(values, bool_list)
            elif isinstance(idx, tuple):
                bool_list.extend([isinstance(it, str) for it in idx])
            else:
                bool_list.extend([False])
            return ReportHeaders._check_list_of_tuples(values, bool_list)



from abc import ABC, abstractmethod


class Validator(ABC):
    def __set_name__(self, owner, name):
        self.private_name = f"_{name}"

    def __get__(self, obj, objtype=None):
        try:
            return getattr(obj, self.private_name)
        except AttributeError:
            return getattr(self, "default_value", None)

    def __set__(self, obj, value) -> None:
        # Some validators may return a value
        value_updated = self.validator(value)
        if value_updated:
            setattr(obj, self.private_name, value_updated)
        else:
            setattr(obj, self.private_name, value)

    @abstractmethod
    def validator(self, value):
        raise NotImplementedError(
            'Descriptors must create a `validator()` class that returns a boolean!'
        )
